@@ -34,7 +34,6 @@ Interrupt and signal handling for Cython
 #include <stdlib.h>
 #include <limits.h>
 #include <errno.h>
-#include <pthread.h>
 #if HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
@@ -54,14 +53,6 @@ Interrupt and signal handling for Cython
 #include <sys/prctl.h>
 #endif
 #include <Python.h>
-#if HAVE_PARI
-#include <pari/pari.h>
-#else
-/* Fake PARI variables */
-static int PARI_SIGINT_block = 0;
-static int PARI_SIGINT_pending = 0;
-#define paricfg_version NULL
-#endif
 
 // Custom signal handling of other packages.
 #define MAX_N_CUSTOM_HANDLERS 16
@@ -94,10 +85,11 @@ void custom_set_pending_signal(int sig){
 }
 
 #if HAVE_WINDOWS_H
-/* We must include <windows.h> after <pari.h>
- * See https://github.com/sagemath/cysignals/issues/107 */
 #include <windows.h>
 #endif
+#if !_WIN32
+#include <pthread.h>
+#endif
 #include "struct_signals.h"
 
 
@@ -118,9 +110,11 @@ static sigset_t default_sigmask;
 static sigset_t sigmask_with_sigint;
 #endif
 
+#if !_WIN32
 /* A trampoline to jump to after handling a signal. */
 static cyjmp_buf trampoline_setup;
 static sigjmp_buf trampoline;
+#endif
 
 static void setup_cysignals_handlers(void);
 static void cysigs_interrupt_handler(int sig);
@@ -204,15 +198,23 @@ static inline void print_stderr_ptr(void *ptr)
 
 /* Reset all signal handlers and the signal mask to their defaults. */
 static inline void sig_reset_defaults(void) {
+#ifdef SIGHUP
     signal(SIGHUP, SIG_DFL);
+#endif
     signal(SIGINT, SIG_DFL);
+#ifdef SIGQUIT
     signal(SIGQUIT, SIG_DFL);
+#endif
     signal(SIGILL, SIG_DFL);
     signal(SIGABRT, SIG_DFL);
     signal(SIGFPE, SIG_DFL);
+#ifdef SIGBUS
     signal(SIGBUS, SIG_DFL);
+#endif
     signal(SIGSEGV, SIG_DFL);
+#ifdef SIGALRM
     signal(SIGALRM, SIG_DFL);
+#endif
     signal(SIGTERM, SIG_DFL);
 #if HAVE_SIGPROCMASK
     sigprocmask(SIG_SETMASK, &default_sigmask, NULL);
@@ -238,10 +240,14 @@ static inline void sigdie_for_sig(int sig, int inside)
             sigdie(sig, "Unhandled SIGFPE during signal handling.");
         else if (sig == SIGSEGV)
             sigdie(sig, "Unhandled SIGSEGV during signal handling.");
+    #ifdef SIGBUS
         else if (sig == SIGBUS)
             sigdie(sig, "Unhandled SIGBUS during signal handling.");
+    #endif
+    #ifdef SIGQUIT
         else if (sig == SIGQUIT)
             sigdie(sig, NULL);
+    #endif
         else
             sigdie(sig, "Unknown signal during signal handling.");
     }
@@ -254,23 +260,112 @@ static inline void sigdie_for_sig(int sig, int inside)
             sigdie(sig, "Unhandled SIGFPE: An unhandled floating point exception occurred.");
         else if (sig == SIGSEGV)
             sigdie(sig, "Unhandled SIGSEGV: A segmentation fault occurred.");
+    #ifdef SIGBUS
         else if (sig == SIGBUS)
             sigdie(sig, "Unhandled SIGBUS: A bus error occurred.");
+    #endif
+    #ifdef SIGQUIT
         else if (sig == SIGQUIT)
             sigdie(sig, NULL);
+    #endif
         else
             sigdie(sig, "Unknown signal received.");
     }
 }
 
+/* Cygwin-specific implementation details */
+#if defined(__CYGWIN__) && defined(__x86_64__)
+#include <w32api/errhandlingapi.h>
+LONG WINAPI win32_altstack_handler(EXCEPTION_POINTERS *exc)
+{
+    int sig = 0;
+    /* If we're not handling a signal there is no reason to execute the
+     * following code; otherwise it can be run in inappropriate contexts
+     * such as when a STATUS_ACCESS_VIOLATION is raised when accessing
+     * uncommitted memory in an mmap created with MAP_NORESERVE. See
+     * discussion at https://trac.sagemath.org/ticket/27214#comment:11
+     *
+     * Unfortunately, when handling an exception that occurred while
+     * handling another signal, there is currently no way (through Cygwin)
+     * to distinguish this case from a legitimate segfault.
+     */
+    if (!cysigs.inside_signal_handler) {
+        return ExceptionContinueExecution;
+    }
 
-/* Additional platform-specific implementation code */
-#if defined(__CYGWIN__)
-#include "implementation_cygwin.c"
+    /* Logic cribbed from Cygwin for mapping common Windows exception
+     * codes to the relevant signal numbers:
+     * https://cygwin.com/git/gitweb.cgi?p=newlib-cygwin.git;a=blob;f=winsup/cygwin/exceptions.cc;h=77eff05707f95f7277974fadbccf0e74223d8d1c;hb=HEAD#l650
+     * Unfortunately there is no external API by which to access this
+     * mapping (a la cygwin_internal(CW_GET_ERRNO_FROM_WINERROR, ...)) */
+    switch (exc->ExceptionRecord->ExceptionCode) {
+        case STATUS_FLOAT_DENORMAL_OPERAND:
+        case STATUS_FLOAT_DIVIDE_BY_ZERO:
+        case STATUS_FLOAT_INVALID_OPERATION:
+        case STATUS_FLOAT_STACK_CHECK:
+        case STATUS_FLOAT_INEXACT_RESULT:
+        case STATUS_FLOAT_OVERFLOW:
+        case STATUS_FLOAT_UNDERFLOW:
+        case STATUS_INTEGER_DIVIDE_BY_ZERO:
+        case STATUS_INTEGER_OVERFLOW:
+            sig = SIGFPE;
+            break;
+        case STATUS_ILLEGAL_INSTRUCTION:
+        case STATUS_PRIVILEGED_INSTRUCTION:
+        case STATUS_NONCONTINUABLE_EXCEPTION:
+            sig = SIGILL;
+            break;
+        case STATUS_TIMEOUT:
+            sig = SIGALRM;
+            break;
+        case STATUS_GUARD_PAGE_VIOLATION:
+        case STATUS_DATATYPE_MISALIGNMENT:
+            sig = SIGBUS;
+            break;
+        case STATUS_ACCESS_VIOLATION:
+            /* In the case of this last resort exception handling we can
+             * probably safely assume this should be a SIGSEGV;
+             * other access violations would have already been handled by
+             * Cygwin before we wound up on the alternate stack */
+        case STATUS_STACK_OVERFLOW:
+        case STATUS_ARRAY_BOUNDS_EXCEEDED:
+        case STATUS_IN_PAGE_ERROR:
+        case STATUS_NO_MEMORY:
+        case STATUS_INVALID_DISPOSITION:
+            sig = SIGSEGV;
+            break;
+        case STATUS_CONTROL_C_EXIT:
+            sig = SIGINT;
+            break;
+    }
+
+    sigdie_for_sig(sig, 1);
+    return ExceptionContinueExecution;
+}
+
+
+static void cygwin_setup_alt_stack() {
+    AddVectoredContinueHandler(0, win32_altstack_handler);
+}
+
+#endif  /* CYGWIN && __x86_64__ */
+
+void get_monotonic_time(struct timespec *ts) {
+#ifdef _WIN32
+    LARGE_INTEGER frequency;
+    LARGE_INTEGER counter;
+
+    QueryPerformanceFrequency(&frequency);
+    QueryPerformanceCounter(&counter);
+
+    ts->tv_sec = counter.QuadPart / frequency.QuadPart;
+    ts->tv_nsec = (counter.QuadPart % frequency.QuadPart) * 1e9 / frequency.QuadPart;
+#else
+    clock_gettime(CLOCK_MONOTONIC, ts);
 #endif
+}
 
-
-/* Handler for SIGHUP, SIGINT, SIGALRM
+/* Handler for SIGHUP, SIGINT, SIGALRM, SIGTERM
  *
  * Inside sig_on() (i.e. when cysigs.sig_on_count is positive), this
  * raises an exception and jumps back to sig_on().
@@ -289,19 +384,21 @@ static void cysigs_interrupt_handler(int sig)
         if (cysigs.debug_level >= 3) print_backtrace();
         /* Store time of this signal, unless there is already a
          * pending signal. */
-        if (!cysigs.interrupt_received) clock_gettime(CLOCK_MONOTONIC, &sigtime);
+        if (!cysigs.interrupt_received) get_monotonic_time(&sigtime);
     }
 #endif
 
     if (cysigs.sig_on_count > 0)
     {
-        if (!cysigs.block_sigint && !PARI_SIGINT_block && !custom_signal_is_blocked())
+        if (!cysigs.block_sigint && !custom_signal_is_blocked())
         {
             /* Raise an exception so Python can see it */
             do_raise_exception(sig);
 
+#if !_WIN32
             /* Jump back to sig_on() (the first one if there is a stack) */
             siglongjmp(trampoline, sig);
+#endif
         }
     }
     else
@@ -315,10 +412,13 @@ static void cysigs_interrupt_handler(int sig)
     /* If we are here, we cannot handle the interrupt immediately, so
      * we store the signal number for later use.  But make sure we
      * don't overwrite a SIGHUP or SIGTERM which we already received. */
-    if (cysigs.interrupt_received != SIGHUP && cysigs.interrupt_received != SIGTERM)
+    if (
+#ifdef SIGHUP
+        cysigs.interrupt_received != SIGHUP && 
+#endif
+        cysigs.interrupt_received != SIGTERM)
     {
         cysigs.interrupt_received = sig;
-        PARI_SIGINT_pending = sig;
         custom_set_pending_signal(sig);
     }
 }
@@ -333,8 +433,11 @@ static void cysigs_signal_handler(int sig)
     int inside = cysigs.inside_signal_handler;
     cysigs.inside_signal_handler = 1;
 
-    if (inside == 0 && cysigs.sig_on_count > 0 && sig != SIGQUIT)
-    {
+    if (inside == 0 && cysigs.sig_on_count > 0 
+        #ifdef SIGQUIT
+            && sig != SIGQUIT
+        #endif
+    ) {
         /* We are inside sig_on(), so we can handle the signal! */
 #if ENABLE_DEBUG_CYSIGNALS
         if (cysigs.debug_level >= 1) {
@@ -342,15 +445,16 @@ static void cysigs_signal_handler(int sig)
             print_stderr_long(sig);
             print_stderr(" *** inside sig_on\n");
             if (cysigs.debug_level >= 3) print_backtrace();
-            clock_gettime(CLOCK_MONOTONIC, &sigtime);
+            get_monotonic_time(&sigtime);
         }
 #endif
 
         /* Raise an exception so Python can see it */
         do_raise_exception(sig);
-
+    #if !_WIN32
         /* Jump back to sig_on() (the first one if there is a stack) */
         siglongjmp(trampoline, sig);
+    #endif
     }
     else
     {
@@ -362,7 +466,7 @@ static void cysigs_signal_handler(int sig)
     }
 }
 
-
+#if !_WIN32
 /* A trampoline to jump to after handling a signal.
  *
  * The jump to sig_on() uses cylongjmp(), which does not restore the
@@ -388,7 +492,7 @@ static void cysigs_signal_handler(int sig)
  * fact, POSIX recommends threads in
  * http://pubs.opengroup.org/onlinepubs/009695299/functions/makecontext.html
  */
-static void* _sig_on_trampoline(void* dummy)
+static void* _sig_on_trampoline(CYTHON_UNUSED void* dummy)
 {
     register int sig;
 
@@ -419,7 +523,7 @@ static void setup_trampoline(void)
     size_t trampolinestacksize = 1 << 17;
 
 #ifdef PTHREAD_STACK_MIN
-    if (trampolinestacksize < PTHREAD_STACK_MIN)
+    if (trampolinestacksize < (size_t) PTHREAD_STACK_MIN)
         trampolinestacksize = PTHREAD_STACK_MIN;
 #endif
     trampolinestack = malloc(trampolinestacksize + 4096);
@@ -446,6 +550,7 @@ static void setup_trampoline(void)
         cylongjmp(trampoline_setup, 1);
     }
 }
+#endif
 
 
 /* This calls sig_raise_exception() to actually raise the exception. */
@@ -454,7 +559,7 @@ static void do_raise_exception(int sig)
 #if ENABLE_DEBUG_CYSIGNALS
     struct timespec raisetime;
     if (cysigs.debug_level >= 2) {
-        clock_gettime(CLOCK_MONOTONIC, &raisetime);
+        get_monotonic_time(&raisetime);
         long delta_ms = (raisetime.tv_sec - sigtime.tv_sec)*1000L + (raisetime.tv_nsec - sigtime.tv_nsec)/1000000L;
         PyGILState_STATE gilstate = PyGILState_Ensure();
         print_stderr("do_raise_exception(sig=");
@@ -486,8 +591,7 @@ static void _sig_on_interrupt_received(void)
     do_raise_exception(cysigs.interrupt_received);
     cysigs.sig_on_count = 0;
     cysigs.interrupt_received = 0;
-    PARI_SIGINT_pending = 0;
-    custom_signal_unblock();
+    custom_set_pending_signal(0);
 
 #if HAVE_SIGPROCMASK
     sigprocmask(SIG_SETMASK, &oldset, NULL);
@@ -499,11 +603,9 @@ static void _sig_on_interrupt_received(void)
 static void _sig_on_recover(void)
 {
     cysigs.block_sigint = 0;
-    PARI_SIGINT_block = 0;
     custom_signal_unblock();
     cysigs.sig_on_count = 0;
     cysigs.interrupt_received = 0;
-    PARI_SIGINT_pending = 0;
     custom_set_pending_signal(0);
 
 #if HAVE_SIGPROCMASK
@@ -534,18 +636,11 @@ static void setup_alt_stack(void)
 #if HAVE_SIGALTSTACK
     /* Space for the alternate signal stack. The size should be
      * of the form MINSIGSTKSZ + constant. The constant is chosen rather
-     * ad hoc but sufficiently large.
-     * MINSIGSTKSZ is no longer a constant starting with glibc 2.34. */
+     * ad hoc but sufficiently large. */
     stack_t ss;
-#if MINSIGSTKSZ_IS_CONSTANT
-    static char alt_stack[MINSIGSTKSZ + 5120 + BACKTRACELEN * sizeof(void*)];
-    ss.ss_sp = alt_stack;
-    ss.ss_size = sizeof(alt_stack);
-#else
     size_t stack_size = MINSIGSTKSZ + 5120 + BACKTRACELEN * sizeof(void*);
     ss.ss_sp = malloc(stack_size);
     ss.ss_size = stack_size;
-#endif
     if (ss.ss_sp == NULL) {perror("cysignals malloc alt signal stack"); exit(1);}
     ss.ss_flags = 0;
     if (sigaltstack(&ss, NULL) == -1) {perror("cysignals sigaltstack"); exit(1);}
@@ -558,6 +653,11 @@ static void setup_alt_stack(void)
 
 static void setup_cysignals_handlers(void)
 {
+#ifdef _WIN32
+    signal(SIGINT, cysigs_interrupt_handler);
+    signal(SIGTERM, cysigs_interrupt_handler);
+    signal(SIGABRT, cysigs_signal_handler);
+#else
     struct sigaction sa;
     memset(&sa, 0, sizeof(sa));
 
@@ -585,21 +685,30 @@ static void setup_cysignals_handlers(void)
     /* Handlers for interrupt-like signals */
     sa.sa_handler = cysigs_interrupt_handler;
     sa.sa_flags = 0;
+#ifdef SIGHUP
     if (sigaction(SIGHUP, &sa, NULL)) {perror("cysignals sigaction"); exit(1);}
+#endif
     if (sigaction(SIGINT, &sa, NULL)) {perror("cysignals sigaction"); exit(1);}
+#ifdef SIGALRM
     if (sigaction(SIGALRM, &sa, NULL)) {perror("cysignals sigaction"); exit(1);}
+#endif
 
     /* Handlers for critical signals */
     sa.sa_handler = cysigs_signal_handler;
     /* Allow signals during signal handling, we have code to deal with
      * this case. */
     sa.sa_flags = SA_NODEFER | SA_ONSTACK;
+#ifdef SIGQUIT
     if (sigaction(SIGQUIT, &sa, NULL)) {perror("cysignals sigaction"); exit(1);}
+#endif
     if (sigaction(SIGILL, &sa, NULL)) {perror("cysignals sigaction"); exit(1);}
     if (sigaction(SIGABRT, &sa, NULL)) {perror("cysignals sigaction"); exit(1);}
     if (sigaction(SIGFPE, &sa, NULL)) {perror("cysignals sigaction"); exit(1);}
+#ifdef SIGBUS
     if (sigaction(SIGBUS, &sa, NULL)) {perror("cysignals sigaction"); exit(1);}
+#endif
     if (sigaction(SIGSEGV, &sa, NULL)) {perror("cysignals sigaction"); exit(1);}
+#endif
 }
 
 
@@ -716,7 +825,6 @@ dienow:
     exit(128 + sig);
 }
 
-
 /* Finally include the macros and inline functions for use in
  * signals.pyx. These require some of the above functions, therefore
  * this include must come at the end of this file. */
